// ==============================================================
//                 ORBITER MODULE: tess
//                  
//
// tess.cpp
// Control module for tess vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double ISP = 3000;
const double RCSTH = 15;
const double MAINTH = 25;
const double ARRAY_SPEED = 0.05;
const double ARRAY_ROT_SPEED = 0.025;
const double FUEL_MASS = 45;
const double EMP_MASS = 320;

#define YELLOW RGB(255, 255, 80)
#define GREEN RGB(0, 225, 0)



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class tess: public VESSEL3 {
public:
	
	tess (OBJHANDLE hVessel, int flightmodel);
	~tess ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void DoKillRot(void);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	int MyID;
	THRUSTER_HANDLE th_rcs[12], th_group[2], th_main;
	PROPELLANT_HANDLE hProp;
	ANIMATIONCOMPONENT_HANDLE Section, Section2, Section3, Section4, Section5, Section6, Section7;

private:
	UINT anim_solp_deploy, anim_solp_rot;
	double solp_deploy_proc, solp_rot_proc;
	int solp_status, kilrot_sw;


};

tess::tess (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{

}

tess::~tess ()
{
	
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void tess::DefineAnimations (void)
{
	// 1. Solar arrays - rotation
	anim_solp_rot = CreateAnimation (0);
	static UINT ArraySpRotGrp1[1] = {0};

	static MGROUP_ROTATE SpRot1 (0, ArraySpRotGrp1, 0, _V(0,0,0), _V(-1,0,0), (float)(PI2));
	Section=AddAnimationComponent (anim_solp_rot, 0.0, 1, &SpRot1);


	
	// 2. Solar arrays - folding
	anim_solp_deploy = CreateAnimation (0);
	static UINT ArrayLFoldGrp1[1] = {12};
	static UINT ArrayLFoldGrp2[1] = {11};
	static UINT ArrayLFoldGrp3[1] = {10};

	static UINT ArrayRFoldGrp1[1] = {7};
	static UINT ArrayRFoldGrp2[1] = {8};
	static UINT ArrayRFoldGrp3[1] = {9};


	static MGROUP_ROTATE ArrayLFold1 (0, ArrayLFoldGrp1, 1, _V(-0.458,0,0), _V(0,0,1), (float)(-PI*0.5));
	Section2=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayLFold1, Section);

	static MGROUP_ROTATE ArrayLFold2 (0, ArrayLFoldGrp2, 1, _V(-0.468,-0.279,0), _V(0,0,1), (float)(PI));
	Section3=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayLFold2, Section2);

	static MGROUP_ROTATE ArrayLFold3 (0, ArrayLFoldGrp3, 1, _V(-0.489,0.276,0), _V(0,0,1), (float)(-PI));
	AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayLFold3, Section3);

	//// STBD
	static MGROUP_ROTATE ArrayRFold1 (0, ArrayRFoldGrp1, 1, _V( 0.458,0,0), _V(0,0,1), (float)(PI*0.5));
	Section4=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayRFold1, Section);

	static MGROUP_ROTATE ArrayRFold2 (0, ArrayRFoldGrp2, 1, _V( 0.468,-0.279,0), _V(0,0,1), (float)(-PI));
	Section5=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayRFold2, Section4);

	static MGROUP_ROTATE ArrayRFold3 (0, ArrayRFoldGrp3, 1, _V( 0.489,0.276,0), _V(0,0,1), (float)(PI));
	AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayRFold3, Section5);


}



// ==============================================================
// tess class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void tess::clbkSetClassCaps (FILEHANDLE cfg)
{
	solp_deploy_proc = 0.0;
	solp_rot_proc = 0.0;
	solp_status = 0;
	kilrot_sw = 0;
	
	DefineAnimations ();
	// physical specs
	SetCW (1, 1, 1, 1);
	SetRotDrag (_V(1,1,1));
	SetPMI (_V(2,2,1));
	SetCrossSections (_V(7,7,4));
	SetSize (2);
	SetEmptyMass (EMP_MASS);
	SetMeshVisibilityMode ((AddMesh ("TESS/tess")), MESHVIS_ALWAYS);
	
	SetCameraOffset (_V(0,1.5,-1.8));
	SetTouchdownPoints (_V(0.1,0.8,-1.8), _V(0.512,-0.853,-1.8), _V(-0.512,-0.853,-1.8));
	SetSurfaceFrictionCoeff (1,1);
	SetDockParams (_V(0,0,-2), _V(0,0,-1), _V(0,1,0));

	// propellant resources

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	
	// ***************** thruster definitions *******************
	th_main = CreateThruster (_V( 0,0,0), _V(0,0, 1), MAINTH, hProp, ISP);
	
	th_rcs[ 0] = CreateThruster (_V( 0.261,-0.106,-0.472), _V(-0.433,-0.75,0.5), RCSTH, hProp, ISP); //pup
	th_rcs[ 1] = CreateThruster (_V(-0.261,-0.106,-0.472), _V( 0.433,-0.75,0.5), RCSTH, hProp, ISP);

	th_rcs[ 2] = CreateThruster (_V( 0.261, 0.106,-0.472), _V(-0.433, 0.75,0.5), RCSTH, hProp, ISP); //pdown
	th_rcs[ 3] = CreateThruster (_V(-0.261, 0.106,-0.472), _V( 0.433, 0.75,0.5), RCSTH, hProp, ISP);

	th_rcs[ 4] = CreateThruster (_V(-0.261,-0.106,-0.472), _V( 0.433,-0.75,0.5), RCSTH, hProp, ISP); // yawr
	th_rcs[ 5] = CreateThruster (_V(-0.261, 0.106,-0.472), _V( 0.433, 0.75,0.5), RCSTH, hProp, ISP);

	th_rcs[ 6] = CreateThruster (_V( 0.261,-0.106,-0.472), _V(-0.433,-0.75,0.5), RCSTH, hProp, ISP); // yawl
	th_rcs[ 7] = CreateThruster (_V( 0.261, 0.106,-0.472), _V(-0.433, 0.75,0.5), RCSTH, hProp, ISP);

	th_rcs[ 8] = CreateThruster (_V(-0.261, 0.106,-0.472), _V( 0.433, 0.75,0.5), RCSTH, hProp, ISP); // bankr
	th_rcs[ 9] = CreateThruster (_V( 0.261,-0.106,-0.472), _V(-0.433,-0.75,0.5), RCSTH, hProp, ISP);

	th_rcs[10] = CreateThruster (_V( 0.261, 0.106,-0.472), _V(-0.433, 0.75,0.5), RCSTH, hProp, ISP); // bankl
	th_rcs[11] = CreateThruster (_V(-0.261,-0.106,-0.472), _V( 0.433,-0.75,0.5), RCSTH, hProp, ISP);

	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	for(int i=0;i<12;i++)
	{
	AddExhaust (th_rcs[i], 0.5, 0.05, (tex_rcs));
	}

	AddExhaust (th_main, 1, 0.1, _V( 0,0,-0.511), _V(0,0,-1), tex_rcs);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[10];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[ 8];
	th_group[1] = th_rcs[ 9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 6];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_main;
	CreateThrusterGroup (th_group, 1, THGROUP_MAIN);



	///////////////////////////////////////////////////////
	/////AirFoils


	CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0.1), VLiftCoeff, 0, 2, 2, 1);
	CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0.1), HLiftCoeff, 0, 2, 2, 1);

	

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new tess (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (tess*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void tess::clbkPostCreation ()
{
	
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\Vessel\\attsustain.wav",	REPLACE_MAIN_THRUST);
	SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);
	
}
void tess::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters

	sprintf (cbuf, "%d %0.4f %0.4f", solp_status, solp_deploy_proc, solp_rot_proc);
	oapiWriteScenario_string (scn, "SLP", cbuf);

	
}
void tess::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "SLP", 3))
		{
            sscanf (line+3, "%d%lf%lf", &solp_status, &solp_deploy_proc, &solp_rot_proc);
		}
		else
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	SetAnimation (anim_solp_deploy, solp_deploy_proc);
	SetAnimation (anim_solp_rot, solp_rot_proc);
	
}






int tess::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{

bool cbSelectTarget(void *id, char *str, void *data);

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
	{
	  switch (key) 
	  {
		case OAPI_KEY_NUMPAD5: 
		if  (kilrot_sw == 0) kilrot_sw = 1;
		else kilrot_sw = 0;
		return 1;
	  }
	  return true;
	} 
  else if (KEYMOD_CONTROL (kstate)) 
	{
  
	}
  else 
	{ // unmodified keys
    switch (key) 
		{
	case OAPI_KEY_G: 
		if (solp_status  == 0)solp_status = 1;
		return 1;

		//case OAPI_KEY_K: 
		//
		//return 1;

		//case OAPI_KEY_J: 
		//
		//return 1;
		}
	}
  
 return 0;
 }


int tess::clbkConsumeDirectKey(char *kstate)
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

if (KEYMOD_SHIFT (kstate))
{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}
return 0;
}


void tess::DoKillRot()
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

		if (drot > abs(avel.x)) avel.x = 0;
		else avel.x = avel.x + (drot * ((abs(avel.x))/-avel.x));
		if (drot > abs(avel.y)) avel.y = 0;
		else avel.y = avel.y + (drot * ((abs(avel.y))/-avel.y));
		if (drot > abs(avel.z)) avel.z = 0;
		else avel.z = avel.z + (drot * ((abs(avel.z))/-avel.z));

		SetAngularVel (avel);
		if (avel.x == 0 && avel.y == 0 && avel.z == 0) kilrot_sw = 0;
}


bool tess::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	//int w5 = (int)(264*d);
	//int w6 = (int)(280*d);
	//int w7 = (int)(296*d);
	//int w8 = (int)(312*d);
	//int w9 = (int)(328*d);
	//int w10 = (int)(344*d);
	//int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);

	{
	skp->SetTextColor(0x0066FF66);
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double mass_fraction = ((m1+m2)/(m1));
	double dv = (ISP*(log(mass_fraction)));
	char cbuf[256];
	sprintf(cbuf,"Current Available dV %0.2f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char hbuf[256];
	sprintf(hbuf,"Shift+NumPad = Reaction Wheel Att.Control");
	skp->Text ((10+roxl), (w1+royl), hbuf,strlen(hbuf));
	}

	if (kilrot_sw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"                                          Kill Rot.Active");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Deploy Solar Panels");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	return true;
}


void tess::clbkPreStep(double SimT, double SimDT, double MJD) {
	//sprintf(oapiDebugString(), "ps fm %d", GetFlightModel());

	if (kilrot_sw == 1) DoKillRot();
	
	///////// animations

	// deploy solp
	if (solp_status == 0) solp_deploy_proc = 0;
	else if (solp_status == 1) 
	{
		double da = SimDT * ARRAY_SPEED;
		solp_deploy_proc = solp_deploy_proc + da;
		if (solp_deploy_proc > 1) 
		{
			solp_deploy_proc = 1;
			solp_status = 2;
		}
	}
	else if (solp_status > 1)
	{
	solp_deploy_proc = 1;
	
	//////// suntracking
	VECTOR3 sun_pos_glob = _V(0,0,0);
	VECTOR3 sun_pos_loc = _V(0,0,0);

	Global2Local(sun_pos_glob, sun_pos_loc);
	double sun_angle = atan2(-sun_pos_loc.z, sun_pos_loc.y);
	if(sun_angle < 0) sun_angle = sun_angle + PI2;

	if (solp_status == 3)sun_angle = PI;
	double sun_angle_proc = sun_angle/PI2;

	int rot_dir = 1;
	double proc_diff = (sun_angle_proc - solp_rot_proc);
	if (proc_diff > 0.5) rot_dir = -1;
	if (proc_diff < 0) rot_dir = -1;
	if (proc_diff < -0.5) rot_dir = 1;

	double da = SimDT * ARRAY_ROT_SPEED;
	if (proc_diff > 0.5)proc_diff = proc_diff - 1; // PI PI2
	else if (proc_diff < -0.5)proc_diff = proc_diff + 1;
	
	if (da > abs(proc_diff)) solp_rot_proc = sun_angle_proc;
	else solp_rot_proc = solp_rot_proc + (da*rot_dir);
	
	int srpi = (int) (solp_rot_proc);
	if (solp_rot_proc > 1) solp_rot_proc = solp_rot_proc - srpi;
	else if (solp_rot_proc < 0) solp_rot_proc = solp_rot_proc + 1 - srpi;
	}
	SetAnimation (anim_solp_rot, solp_rot_proc);
	SetAnimation (anim_solp_deploy, solp_deploy_proc);
	
}


//void tess::clbkPostStep (double simt, double simdt, double mjd)
//{
//
//}



